/*
    foo2kyo.c - foomatic filter for Kyocera Mita FS-1016MFP
    Copyright (C) 2006, Florian Demski

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*
 *  LEGAL NOTICE concerning jbigkit used by foo2kyo:
 *
 *  It is possible that certain products which can be built using this
 *  software module might form inventions protected by patent rights in
 *  some countries (e.g., by patents about arithmetic coding algorithms
 *  owned by IBM and AT&T in the USA). Provision of this software by the
 *  author does NOT include any licences for any patents. In those
 *  countries where a patent licence is required for certain applications
 *  of this software module, you will have to obtain such a licence
 *  yourself.
 */




#include<stdio.h>
#include<stdlib.h>
#include<ctype.h>
#include<string.h>
#include<unistd.h>
#include<stdarg.h>

#include "jbig.h"

static char Version[] = "foo2kyo.c 0.1.0";

int Debug=0;
int PaperCode=0;
int MediaCode=0;
int PageWidth=4768;
int PageHeight=6816;
int Copies=1;
int Brightness=4;
int Eco=0;

int current_page=0;

struct data_node {
	int size;
	unsigned char *d;
	struct data_node *next;
};

const char pjl_start[] = 
    "\x1B\%-12345X@PJL\r\n"
    "@PJL ENTER LANGUAGE=PL\r\n"
    "\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0"
    "\x0\x0\x0\x0\x0\x0\x0\x0\x0";
    
const char pjl_end[] = 
    "\x1B\%-12345X@PJL\r\n"
    "@PJL EOJ\r\n"
    "\x1B\%-12345X"
    "\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0"
    "\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0";

unsigned char start_block[64] = {
	0x0B, 0x06, 0x11, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char eco_block[64] = {
	0x0B, 0x05, 0x19, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char brightness_block[64] = {
	0x0B, 0x05, 0x15, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char page1_block[64] = {
	0x0B, 0x12, 0x12, 0x00, 0x00, 0x3C, 0x3C, 0x01, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0xFF, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char page2_block[64] = {
	0x0B, 0x0C, 0x13, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char page3_block[64] = {
	0x0B, 0x04, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char page4_block[64] = {
	0x0B, 0x12, 0x0C, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char cs1_block[64] = {
	0x0B, 0x06, 0x1A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char cs2_block[64] = {
	0x0B, 0x08, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char pageend_block[64] = {
	0x0B, 0x04, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char end1_block[64] = {
	0x0B, 0x04, 0x33, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};
unsigned char end2_block[64] = {
	0x0B, 0x04, 0x17, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};


const unsigned char bit_table[256] =  {
	0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
	1, 2, 3, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	1, 2, 3, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 4, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	1, 2, 3, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 4, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	2, 3, 4, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 5, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	1, 2, 3, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 4, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	2, 3, 4, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 5, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	2, 3, 4, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 5, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	3, 4, 5, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	4, 5, 6, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8 
};

const unsigned char mirror_table[256] = {
	0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0, 
	0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8, 
	0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4, 
	0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC, 
	0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2, 
	0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA, 
	0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6, 
	0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE, 
	0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1, 
	0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9, 
	0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5, 
	0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD, 
	0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3, 
	0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB, 
	0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7, 
	0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
};

void usage(void)
{
	fprintf(stderr,
		"Usage:\n"
		"    foo2kyo [options] <pbmraw-file >kyo-file\n"
		"\n"
		"        Convert Ghostscript pbmraw format to a stream for driving\n"
		"        the Kyocera Mita FS-1016MFP laser printer\n"
		"        and possibly other Kyocera Mita printers.\n"
		"\n"
		"Options:\n"
		"    -g <xpix>x<ypix> Set page dimensions in pixels [%dx%d]\n"
		"    -p <paper code>  Paper code to send to printer [%d]\n"
        "        0=A4, 1=letter, 2=legal, 3=A5, 4=B5(JIS),\n"
        "        5=oficioII, 6=executive, 7=statement,\n"
        "        8=folio, 9=B6(JIS), 10=A6, 11=B5(ISO),\n"
        "        12=envC5, 13=env#10, 14=envCL, 15=env#9,\n"
        "        16=envMonarch, 17=env#6, 18=16K\n"
		"    -m <media code>  Media code to send to printer [%d]\n"
        "        0=plainPaper, 1=transparency, 2=preprinted,\n"
        "        3=labels, 4=bond, 5=recycled, 6=vellum,\n"
        "        7=rough, 8=letterhead, 9=color, 10=prepunched,\n"
        "        11=envelope, 12=cardStock, 13=thickPaper,\n"
        "        14=highQuality, 15=customType\n"
		"    -n <copies>      Number of copies [%d]\n"
		"    -b <brightness>  Set brightness 1=light ... 7=dark [%d]\n"
		"    -e <eco>         Set Eco-mode [%d]\n"
        "Debugging Options:\n"
		"    -D <debug level> Set Debug Level [%d]\n" 
		"    -V               Version %s\n",
		PageWidth, PageHeight,
		PaperCode,
		MediaCode,
		Copies,
		Brightness,
		Eco,
		Debug,
		Version);
	exit(1);
}

void debug(int level, char *fmt, ...)
{
	va_list ap;
	
	if (Debug < level)
		return;
	
	setvbuf(stderr, (char *)NULL, _IOLBF, BUFSIZ);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}

void error(int fatal, char *fmt, ...)
{
	va_list ap;
	
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	
	if (fatal)
	    exit(fatal);
}

int parse_xy(char *str, int *xp, int *yp)
{
	char *p;

    if (!str || str[0] == 0) return -1;

    *xp = strtoul(str, &p, 0);
    if (str == p) return -2;
    while (*p && (*p < '0' || *p > '9'))
	++p;
    str = p;
    if (str[0] == 0) return -3;
    *yp = strtoul(str, &p, 0);
    if (str == p) return -4;
    return (0);
}

void do_checksum(unsigned char *block)
{
	unsigned char cs;
	int i;
	
	cs = 0;
	
	for (i=0; i<block[1]-1; i++)
	    cs += block[i];
	    
	cs = (unsigned char)(0x100 - cs);
	block[block[1]-1] = cs;
}

void start_doc(void)
{
	fwrite(pjl_start, 1, 0x40, stdout);
	
	do_checksum(start_block);
	fwrite(start_block, 1, 0x40, stdout);
	
	eco_block[3] = (Eco) ? 1 : 0;
	do_checksum(eco_block);
	fwrite(eco_block, 1, 0x40, stdout);
	
	brightness_block[3] = (unsigned char) Brightness;
	do_checksum(brightness_block);
	fwrite(brightness_block, 1, 0x40, stdout);
}

void end_doc(void)
{
	do_checksum(end1_block);
	fwrite(end1_block, 1, 0x40, stdout);
	
	do_checksum(end2_block);
	fwrite(end2_block, 1, 0x40, stdout);
	
	fwrite(pjl_end, 1, 0x40, stdout);	
}

void skip_to_nl(void)
{
	int c;
	
	do {
		c = fgetc(stdin);
	} while (c != 0x0A);
}

void _data_out(unsigned char *buf, size_t len, void *file)
{
	struct data_node *p = (struct data_node *)file;
	
	if (p->size == 0) {
		p->d = (unsigned char*) malloc(len);
		memcpy(p->d, buf, len);
		p->size = len;
	} else {
		while (p->next) p = p->next;
		p->next = malloc(sizeof(struct data_node));
		p = p->next;

		p->d = (unsigned char*) malloc(len);
		memcpy(p->d, buf, len);
		p->size = len;
		p->next = NULL;		
	}
}

int next_page(void)
{
	int c1, c2, c3, i, j;
	int xs=0, ys=0;
	int in_xs=1;
	int row_len;
	unsigned bits_set, sz_jbg, cs;
	unsigned char *pic_buf;
	struct jbg_enc_state jbg;
	struct data_node root, *pnode;
		
	c1 = fgetc(stdin);
	c2 = fgetc(stdin);
	c3 = fgetc(stdin);
	
	if ((c1 != 'P') || (c2 != '4') || (c3 != 0x0A)) 
		return 0;	
		
	/* header found --- check for information */
	for(;;) {
		c1 = fgetc(stdin);
		
		if (c1 == '#') {
			skip_to_nl();
			continue;
		}
		
		if (in_xs == 1) {
			if (c1 != ' ') {
				xs *= 10;
				xs += c1 - '0';
			} else {
				in_xs++;
			}
		} else if (in_xs==2) {
			if (c1 != 0x0A) {
				ys *= 10;
				ys += c1 - '0';
			} else {
				break;
			}
		}		
	}
	
	current_page++;

	root.d = NULL;
	root.next = NULL;
        root.size = 0;

	row_len = (xs + 7) / 8;
	debug(1, "next_page: %d x %d --- row_len=%d\n", xs, ys, row_len);
	
	pic_buf = (unsigned char*)malloc(row_len * ys);
	if (pic_buf == NULL) {
		error(-1, "Error allocating %d bytes.\n", row_len * ys);
		return 0;
	}

	if (fread(pic_buf, row_len, ys, stdin) != ys) {
		free(pic_buf);
		error(-2, "Error reading stdin.\n");
		return 0;
	}
	
	/* mirror the buffer and calculate number of bits set*/
	bits_set = 0;
	for (i=0; i<ys; i++) {
		unsigned char *ls, *le;
		
		ls = pic_buf + i * row_len;
		le = ls + row_len - 1;
		
		while (le > ls) {
			unsigned char tmp;
			
			tmp = *ls;
			*ls = mirror_table[*le];
			*le = mirror_table[tmp];
	
			bits_set += bit_table[*ls];
			bits_set += bit_table[*le];
			
			ls++;
			le--;
		}
	}
	
	/* generate JBIG data */
	jbg_enc_init(&jbg, xs, ys, 1, &pic_buf, _data_out, &root);
	jbg_enc_layers(&jbg, 0);
	jbg_enc_options(&jbg, 2, JBG_TPDON | JBG_TPBON | JBG_DPON, ys, 0, 0);
	jbg_enc_out(&jbg);
	jbg_enc_free(&jbg);
	
	/* calculate size and checksum */
	/* leave out the 20 byte long header */
	
	sz_jbg = 0;
	cs = 0;
	
	pnode = root.next;
	while (pnode) {
		sz_jbg += pnode->size;
		
		for (i=0; i<pnode->size; i++)
			cs += pnode->d[i];
		
		pnode = pnode->next;
	}
	

	page1_block[3] = (unsigned char) PaperCode;
	page1_block[4] = (unsigned char) MediaCode;
	page1_block[9] = (unsigned char) Copies;
	page1_block[15] = (unsigned char) current_page;
	
	do_checksum(page1_block);
	fwrite(page1_block, 1, 0x40, stdout);

	do_checksum(page2_block);
	fwrite(page2_block, 1, 0x40, stdout);

	do_checksum(page3_block);
	fwrite(page3_block, 1, 0x40, stdout);


	page4_block[5] = (unsigned char)(xs & 0xFF);
	page4_block[6] = (unsigned char)(xs >> 8);
	page4_block[7] = (unsigned char)(ys & 0xFF);
	page4_block[8] = (unsigned char)(ys >> 8);
	page4_block[9] = (unsigned char)(xs & 0xFF);
	page4_block[10] = (unsigned char)(xs >> 8);
	page4_block[11] = (unsigned char)(ys & 0xFF);
	page4_block[12] = (unsigned char)(ys >> 8);
	page4_block[13] = (unsigned char)(sz_jbg & 0xFF);
	page4_block[14] = (unsigned char)((sz_jbg >> 8) & 0xFF);
	page4_block[15] = (unsigned char)((sz_jbg >> 16) & 0xFF);
	page4_block[16] = (unsigned char)((sz_jbg >> 24) & 0xFF);
	
	do_checksum(page4_block);
	fwrite(page4_block, 1, 0x40, stdout);


	/* output data */
	
	pnode = root.next;
	while (pnode) {
		fwrite(pnode->d, 1, pnode->size, stdout);
		
		pnode = pnode->next;
	}


	cs1_block[3] = (unsigned char)(cs & 0xFF);
	cs1_block[4] = (unsigned char)(cs >> 8);
	
	do_checksum(cs1_block);
	fwrite(cs1_block, 1, 0x40, stdout);
	
	cs2_block[3] = (unsigned char)(bits_set & 0xFF);
	cs2_block[4] = (unsigned char)((bits_set>>8) & 0xFF);
	cs2_block[5] = (unsigned char)((bits_set>>16) & 0xFF);
	cs2_block[6] = (unsigned char)((bits_set>>24) & 0xFF);
	
	do_checksum(cs2_block);
	fwrite(cs2_block, 1, 0x40, stdout);

	do_checksum(pageend_block);
	fwrite(pageend_block, 1, 0x40, stdout);

	
	/* free data structures */
	if (root.d) 
		free(root.d);
	
	pnode = root.next;
	while (pnode) {
		struct data_node *tmp;
		
		tmp = pnode;
		
		if (pnode->d)
			free(pnode->d);
			
		pnode = pnode->next;
		free(tmp);
	}
	
	free(pic_buf);
	
	return 1;
}

int main(int argc, char *argv[])
{
	int c, i, j;
	
	while((c=getopt(argc, argv, "g:p:m:n:b:e:D:V?h")) != EOF) {
		switch(c) {
		case 'g':
			if (parse_xy(optarg, &PageWidth, &PageHeight))
			    error(1, "Illegal format '%s' for -g\n", optarg);
			 break;
	    case 'm':
	    	MediaCode = atoi(optarg);
	    	break;
	    case 'n':
	    	Copies = atoi(optarg);
	    	break;
	    case 'p':
	    	PaperCode = atoi(optarg);
	    	break;
	    case 'b':
	    	Brightness = atoi(optarg);
	    	break;
	    case 'e':
	    	Eco = atoi(optarg);
	    	break;
	    case 'D':
	    	Debug = atoi(optarg);
	    	break;
	    case 'V':
	    	printf("%s\n", Version);
	    	exit(0);
	    	break;
	    default:
	        usage();
	        exit(1);
	        break;
	    }
	}
	
	start_doc();

	for (;;) {
		if (!next_page())
			break;
	}
	
	end_doc();
}



